@@ -12,12 +12,13 @@ from requests import Response
 from web3 import Web3, HTTPProvider
 from web3.types import RPCEndpoint, RPCResponse, TxParams
 
+from eulith_web3.common import INT_FEE_TO_FLOAT_DIVISOR
 from eulith_web3.erc20 import TokenSymbol, EulithWETH, EulithERC20
 from eulith_web3.exceptions import EulithAuthException, EulithRpcException
-from eulith_web3.requests import EulithShortOnRequest, EulithShortOffRequest, EulithAaveV2StartLoanRequest
+from eulith_web3.requests import EulithShortOnRequest, EulithShortOffRequest, EulithAaveV2StartLoanRequest, FlashRequest
 from eulith_web3.swap import EulithSwapProvider, EulithSwapRequest, EulithLiquiditySource
 from eulith_web3.uniswap import EulithUniswapV3Pool, EulithUniV3StartLoanRequest, EulithUniV3StartSwapRequest, \
-    EulithUniV3SwapQuoteRequest, EulithUniswapPoolLookupRequest, UniswapV3PoolFee
+    EulithUniV3SwapQuoteRequest, EulithUniswapPoolLookupRequest, UniswapPoolFee
 
 
 class ApiToken:
@@ -83,7 +84,8 @@ class EulithData:
         self.api_access_token: ApiToken = get_api_access_token(self.eulith_url, self.eulith_refresh_token)
         headers = get_headers(eulith_url, self.api_access_token.token)
         self.http = HTTPProvider(endpoint_uri=eulith_url, request_kwargs={
-            'headers': headers
+            'headers': headers,
+            'timeout': 60
         })
 
     def send_transaction(self, params) -> RPCResponse:
@@ -225,7 +227,7 @@ class EulithData:
     def start_uniswap_v3_swap(self, params: EulithUniV3StartSwapRequest) -> (bool, RPCResponse):
         try:
             sell_token = params.get('sell_token').address
-            sell_amount = params.get('sell_amount')
+            amount = params.get('amount')
             pool_address = params.get('pool_address')
             fill_or_kill = params.get('fill_or_kill')
             sqrt_limit_price = params.get('sqrt_limit_price')
@@ -234,7 +236,7 @@ class EulithData:
 
             parsed_params = {
                 'sell_token': sell_token,
-                'sell_amount': sell_amount,
+                'amount': amount,
                 'pool_address': pool_address,
                 'fill_or_kill': fill_or_kill,
                 'sqrt_limit_price': sqrt_limit_price
@@ -281,7 +283,8 @@ class EulithData:
             parsed_params = {
                 'sell_token': params.get('sell_token').address,
                 'buy_token': params.get('buy_token').address,
-                'sell_amount': params.get('sell_amount')
+                'amount': params.get('amount'),
+                'true_for_amount_in': params.get('true_for_amount_in', True)
             }
 
             fee = params.get('fee', None)
@@ -443,30 +446,31 @@ class EulithWeb3(Web3):
         else:
             raise EulithRpcException(contract_address_or_error)
 
-    def parse_uni_quote_to_swap_request(
-            self,
-            res,
-            fill_or_kill: bool,
-            recipient: Optional[ChecksumAddress],
-            pay_transfer_from: Optional[ChecksumAddress]
-    ) -> (float, EulithUniV3StartSwapRequest):
+    # returns price (float), fee (UniswapPoolFee), swap_request (EulithUniV3StartSwapRequest)
+    def parse_uni_quote_to_swap_request(self, res, fill_or_kill: bool, recipient: Optional[ChecksumAddress],
+                                        pay_transfer_from: Optional[ChecksumAddress]) -> (float, UniswapPoolFee, EulithUniV3StartSwapRequest):
         result = res.get('result')
         price = result.get('price')
         sell_token_address = result.get('sell_token')
         sell_token = EulithERC20(self, sell_token_address)
-        sell_amount = result.get('sell_amount')
+        amount = result.get('amount')
         pool_address = result.get('pool_address')
         limit_price = result.get('sqrt_limit_price')
+        fee = result.get('fee')
+        true_for_amount_in = result.get('true_for_amount_in')
+
+        if not true_for_amount_in:
+            amount *= -1.0  # make negative if we want exact amount out
 
         swap_request = EulithUniV3StartSwapRequest(sell_token=sell_token,
-                                                   sell_amount=sell_amount,
+                                                   amount=amount,
                                                    pool_address=self.toChecksumAddress(pool_address),
                                                    fill_or_kill=fill_or_kill,
                                                    sqrt_limit_price=limit_price,
                                                    recipient=recipient,
                                                    pay_transfer_from=pay_transfer_from)
 
-        return price, swap_request
+        return price, UniswapPoolFee(fee), swap_request
 
 
 # Namespace
@@ -486,13 +490,13 @@ class v0:
     def rollback_atomic_transaction(self):
         return self.ew3.eulith_rollback_transaction()
 
-    def get_proxy_address(self, account: str) -> str:
+    def get_toolkit_contract_address(self, account: str) -> str:
         return self.ew3.eulith_contract_address(account)
 
-    def ensure_proxy_exists(self, account: str) -> str:
+    def ensure_toolkit_contract(self, account: str) -> str:
         return self.ew3.eulith_create_contract_if_not_exist(account)
 
-    def create_proxy(self, account: str) -> str:
+    def create_toolkit_contract(self, account: str) -> str:
         return self.ew3.eulith_create_contract(account)
 
     def refresh_api_token(self):
@@ -535,41 +539,50 @@ class v0:
             token_one = inner_result.get('token_one')
             fee = inner_result.get('fee')
             pool_address = inner_result.get('pool_address')
-            return EulithUniswapV3Pool(self.ew3, self.ew3.toChecksumAddress(pool_address), UniswapV3PoolFee(fee),
+            return EulithUniswapV3Pool(self.ew3, self.ew3.toChecksumAddress(pool_address), UniswapPoolFee(fee),
                                        self.ew3.toChecksumAddress(token_zero), self.ew3.toChecksumAddress(token_one))
         else:
             raise EulithRpcException(res)
 
-    def start_flash_loan(self, params: Union[EulithUniV3StartLoanRequest, EulithAaveV2StartLoanRequest]) -> bool:
+    # returns fee (float) as percent: i.e. 0.001 = 0.1%
+    def start_flash_loan(self, params: Union[EulithUniV3StartLoanRequest, EulithAaveV2StartLoanRequest]) -> float:
         param_keys = params.keys()
         if 'borrow_token_a' in param_keys:
             status, res = self.ew3.eulith_data.start_uniswap_v3_loan(params)
         else:
             status, res = self.ew3.eulith_data.start_aave_v2_loan(params)
 
-        return status
+        if status:
+            fee = int(res.get('result'), 16)
+            return fee / INT_FEE_TO_FLOAT_DIVISOR
+        else:
+            raise EulithRpcException(res)
 
-    def start_swap(self, params: EulithUniV3StartSwapRequest) -> bool:
+    # returns fee (float) as percent: i.e. 0.001 = 0.1%
+    def start_uni_swap(self, params: EulithUniV3StartSwapRequest) -> float:
         status, res = self.ew3.eulith_data.start_uniswap_v3_swap(params)
-        return status
+        if status:
+            fee = int(res.get('result'), 16)
+            return fee / INT_FEE_TO_FLOAT_DIVISOR
+        else:
+            raise EulithRpcException(res)
 
-    def get_univ3_best_price_quote(
-            self,
-            sell_token: EulithERC20,
-            buy_token: EulithERC20,
-            sell_amount: float,
-            fill_or_kill: Optional[bool] = True,
-            recipient: Optional[ChecksumAddress] = None,
-            pay_transfer_from: Optional[ChecksumAddress] = None
-    ) -> (float, EulithUniV3StartSwapRequest):
+    # returns price (float), fee (float), swap_request (EulithUniV3StartSwapRequest)
+    def get_univ3_best_price_quote(self, sell_token: EulithERC20, buy_token: EulithERC20, amount: float,
+                                   true_for_amount_in: Optional[bool] = True, fill_or_kill: Optional[bool] = True,
+                                   recipient: Optional[ChecksumAddress] = None, pay_transfer_from: Optional[ChecksumAddress] = None) -> (float, float, EulithUniV3StartSwapRequest):
         parsed_params = EulithUniV3SwapQuoteRequest(
             sell_token=sell_token,
             buy_token=buy_token,
-            sell_amount=sell_amount,
+            amount=amount,
+            true_for_amount_in=true_for_amount_in,
             fee=None)
         status, res = self.ew3.eulith_data.uniswap_v3_quote(parsed_params)
         if status:
-            return self.ew3.parse_uni_quote_to_swap_request(res, fill_or_kill, recipient, pay_transfer_from)
+            price, fee, swap_request = self.ew3.parse_uni_quote_to_swap_request(res, fill_or_kill, recipient, pay_transfer_from)
+            return price, fee / INT_FEE_TO_FLOAT_DIVISOR, swap_request
+        else:
+            raise EulithRpcException(res)
 
     def finish_inner(self) -> int:
         status, res = self.ew3.eulith_data.finish_inner()
@@ -578,6 +591,34 @@ class v0:
         else:
             raise EulithRpcException(res)
 
+    # returns price (float), fee (float) as percent: i.e. 0.001 = 0.1%
+    def start_flash(self, params: FlashRequest) -> (float, float):
+        amount = params.get('take_amount')
+        pay_transfer_from = params.get('pay_transfer_from', None)
+        recipient = params.get('recipient', None)
+
+        if params.get('take').address.lower() == params.get('pay').address.lower():
+            fee = self.start_flash_loan(EulithUniV3StartLoanRequest(
+                borrow_token_a=params.get('take'),
+                borrow_amount_a=amount,
+                borrow_token_b=None,
+                borrow_amount_b=None,
+                pay_transfer_from=pay_transfer_from,
+                recipient=recipient
+            ))
+
+            return 1.0, fee
+        else:
+            price, fee, swap_request = self.get_univ3_best_price_quote(params.get('pay'),
+                                                                       params.get('take'),
+                                                                       amount,
+                                                                       False, True, recipient, pay_transfer_from)
+            fee = self.start_uni_swap(swap_request)
+            return price, fee
+
+    def pay_flash(self) -> int:
+        return self.finish_inner()
+
 
 def eulith_atomic_middleware(
         make_request: Callable[[RPCEndpoint, Any], Any], web3: "Web3"
